/*
 Vitaly Checkryzhev, <virty1k@mail.ru>
 WWW: http://phoenix.strchat.ru
 ICQ: 55371440
 JID: vcheckryzhev@jabber.ru
*/

#include <mpi.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

// Password length
int seqlen;
unsigned char *seq;
char exit_flag = 0;		// Handle Ctrl-C

#ifdef S_MD5
#include <openssl/md5.h>
#define DIGEST_LENGTH MD5_DIGEST_LENGTH
#else
#include <openssl/sha.h>
#define DIGEST_LENGTH SHA_DIGEST_LENGTH
#endif

#define tablen 62
const char chartable[tablen] =
 "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
#define lastseqfile "lastseq.txt"

#include "sequence.c"

// Queue len
long int queuelen;

 /*   ,     , 
         */
void fsignal (int sig)
{
 switch (sig)
   {
   case SIGALRM:
    exit_flag = 0;
    break;
   case SIGTERM:
   case SIGINT:
   case SIGQUIT:
    // Some exit stuff
    if (exit_flag)
     exit (0);
    alarm (1);
    exit_flag = 1;
    break;
   default:
    break;
   }
}

int main (int argc, char **argv)
{
 double t;
 int i,
   flag,
   size,
   rank;
 MPI_Status stat;
 MPI_Request *req;
 unsigned char testhash[DIGEST_LENGTH * 2],
   stxt[seqlen],
   hash[16],
   chash[16],
   found;
 unsigned long int c,
   q;
 char *wstate,
  *fres;

 MPI_Init (&argc, &argv);
 MPI_Comm_size (MPI_COMM_WORLD, &size);
 MPI_Comm_rank (MPI_COMM_WORLD, &rank);

 if (rank == 0)
   {
    if (argc < 3)
      {
#ifdef S_MD5
       printf ("Compiled with MD5");
#else
       printf ("Compiled with SHA1");
#endif
       printf ("\nUsage: %s password_len hash [seq_per_process]\n", argv[0]);
       MPI_Abort (MPI_COMM_WORLD, 2);
       return 1;
      }

    if (size < 2)
      {
       printf ("Error: %s should run on at least 2 processes\n", argv[0]);
       MPI_Abort (MPI_COMM_WORLD, 1);
       return 2;
      }

    seqlen = atoi (argv[1]);
    if (seqlen < 1)
      {
       printf ("Password length must be greater than 0\n");
       MPI_Abort (MPI_COMM_WORLD, 3);
       return 1;
      }

    strncpy (testhash, argv[2], DIGEST_LENGTH * 2);

    queuelen = atoi (argv[3]);
    if (queuelen < 1)
     queuelen = 100000;


    for (i = 0; i < DIGEST_LENGTH; i++)
     hash[i] = hexsym (&testhash[i * 2]);
   }

 MPI_Bcast (&seqlen, 1, MPI_INT, 0, MPI_COMM_WORLD);
 MPI_Bcast (&hash, DIGEST_LENGTH, MPI_CHAR, 0, MPI_COMM_WORLD);
 seq = malloc (seqlen * sizeof (char));

 if (rank == 0)
   {
    for (i = 1; i < 32; i++)	/*      fsignal */
     signal (i, fsignal);

    fres = malloc ((size - 1) * sizeof (char));
    wstate = malloc ((size - 1) * sizeof (char));
    req = malloc ((size - 1) * sizeof (MPI_Request));

    // set all state to free
    for (i = 0; i < size - 1; i++)
      {
       wstate[i] = 0;
       fres[i] = 0;
      }

    // Init first sequence
    for (i = 0; i < seqlen; i++)
     seq[i] = 0;

    printf ("Run on %d processes\n", size);

    // Iterate all possible sequences
    t = MPI_Wtime ();

    i = 0;
    found = 0;
    while (seq[0] < tablen)
      {
       if (wstate[i] == 0)
	 {
	  // Give new portion
	  wstate[i] = 1;

	  MPI_Send (&queuelen, 1, MPI_LONG_INT, i + 1, 2, MPI_COMM_WORLD);
	  MPI_Send (seq, seqlen, MPI_CHAR, i + 1, 1, MPI_COMM_WORLD);
	  MPI_Irecv (&fres[i], 1, MPI_CHAR, i + 1, 3, MPI_COMM_WORLD,
		     &req[i]);

	  add_seq (seq, seqlen, queuelen);
	 }
       else if (wstate[i] < 10)
	 {
	  // Check if we have an answer
	  MPI_Test (&req[i], &flag, &stat);
	  if (flag)
	    {
	     wstate[i] = 0;
	     if (fres[i])
	       {
		found = 1;
		break;
	       }
	    }
	  else
	   wstate[i]++;
	 }
       else if (wstate[i] == 10)
	 {
	  wstate[i] = 1;
	  sleep (1);
	 }

       i++;
       if (i == size - 1)
	i = 0;
      }

    c = 0;
    for (i = 0; i < size - 1; i++)
     if (fres[i] == 0)
      MPI_Send (&c, 1, MPI_LONG_INT, i + 1, 2, MPI_COMM_WORLD);

    if (found == 0)
     printf ("Shit! Sequence not found.\n");

    t = MPI_Wtime () - t;
    printf ("Time elapsed: %.5f sec.\n", t);

    free (req);
    free (fres);
    free (wstate);
   }
 else
  while (1)
    {
     MPI_Recv (&q, 1, MPI_LONG_INT, 0, 2, MPI_COMM_WORLD, &stat);
     // No more work to do
     if (q == 0)
      break;

     MPI_Recv (seq, seqlen, MPI_CHAR, 0, 1, MPI_COMM_WORLD, &stat);

     printf ("[%d]: ", rank);
     print_seq (seq, seqlen);

     c = 0;
     found = 0;
     while (c < q)
       {
	// Get text presentation
	for (i = 0; i < seqlen; i++)
	 stxt[i] = chartable[seq[i]];

#ifdef S_MD5
	MD5 (stxt, seqlen, chash);
#else
	SHA1 (stxt, seqlen, chash);
#endif

	if (strncmp (hash, chash, seqlen) == 0)
	  {
	   printf ("Found: ");
	   print_seq (seq, seqlen);
	   found = 1;
	   break;
	  }

	step_seq (seq, seqlen);
	c++;
       }

     MPI_Send (&found, 1, MPI_CHAR, 0, 3, MPI_COMM_WORLD);
     if (found == 1)
      break;
    }

 free (seq);
 MPI_Barrier (MPI_COMM_WORLD);
 MPI_Finalize ();
 return 0;
}
